/**
  *  \file client/chartmode.cc
  *  \brief Starchart Mode Handling
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/chartmode.h"
#include "cpluslib/assert.h"

/******************************* WChartMode ******************************/

/** Constructor. */
WChartMode::WChartMode(int kind)
    : parent(0),
      next(0),
      kind(kind)
{ }

/** Destructor. */
WChartMode::~WChartMode()
{
    if (parent)
        parent->removeMode(*this);
}

/** Get next mode in stacking hierarchy. */
WChartMode*
WChartMode::getNextMode() const
{
    return next;
}

/** Get parent WChartModeList. */
WChartModeList*
WChartMode::getParent()
{
    return parent;
}

/**************************** WSimpleChartMode ***************************/

/** Constructor. */
WSimpleChartMode::WSimpleChartMode(int kind)
    : WChartMode(kind)
{ }

/** Destructor. */
WSimpleChartMode::~WSimpleChartMode()
{ }

void
WSimpleChartMode::onMove(GChartLocation& loc)
{
    if (WChartMode* m = getNextMode())
        m->onMove(loc);
}

void
WSimpleChartMode::drawBelow(GfxCanvas& can, GChartLocation& loc)
{
    if (WChartMode* m = getNextMode())
        m->drawBelow(can, loc);
}

void
WSimpleChartMode::drawCursor(GfxCanvas& can, GChartLocation& loc)
{
    if (WChartMode* m = getNextMode())
        m->drawCursor(can, loc);
}

void
WSimpleChartMode::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    if (WChartMode* m = getNextMode())
        m->drawOverlays(can, loc);
}

bool
WSimpleChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    WChartMode* m = getNextMode();
    return m && m->handleEvent(event, loc);
}

/***************************** WChartModeList ****************************/

/** Constructor. */
WChartModeList::WChartModeList()
    : current_mode(0),
      changed(false)
{ }

/** Destructor. */
WChartModeList::~WChartModeList()
{
    deleteAllModes();
}

/** Delete all modes. This assumes all modes are allocated on heap, of course. */
void
WChartModeList::deleteAllModes()
{
    while (current_mode != 0)
        delete current_mode;
}

/** Add new mode. WChartModeList assumes ownership of this object
    unless your delete it before giving WChartModeList a chance to do
    so (i.e. calling deleteAllModes() or destroying the WChartModeList).

    If there is another mode active that has the same kind as \c
    which, that one will be deleted. */
void
WChartModeList::pushMode(WChartMode& which)
{
    ASSERT(which.parent == 0);
    ASSERT(which.next == 0);
    ASSERT(current_mode != &which);

    /* Find place to insert this */
    WChartMode** p = &current_mode;
    while (*p != 0 && (*p)->kind < which.kind)
        p = &(*p)->next;

    /* Insert */
    which.parent = this;
    which.next = *p;
    *p = &which;
    markChanged();

    /* Remove next if it is of the same kind */
    if (which.next != 0 && which.next->kind == which.kind)
        delete which.next;
}

/** Get mode by kind. If there is a mode whose kind is k, return that;
    otherwise return 0. */
WChartMode*
WChartModeList::getModeByKind(int k) const
{
    WChartMode* p = current_mode;
    while (p != 0 && p->kind != k)
        p = p->next;
    return p;
}

/** Remove a mode. Does not delete it, though. */
void
WChartModeList::removeMode(WChartMode& which)
{
    ASSERT(which.parent == this);
    WChartMode** base = &current_mode;
    while (*base != 0) {
        if (*base == &which) {
            // We're about to remove this thing
            *base = which.next;
        } else {
            base = &(*base)->next;
        }
    }
    which.parent = 0;
    markChanged();
}

/** Get current mode. */
WChartMode*
WChartModeList::getCurrentMode() const
{
    return current_mode;
}

/** Handle movement in starchart. */
void
WChartModeList::onMove(GChartLocation& loc)
{
    if (current_mode != 0)
        current_mode->onMove(loc);
}

/** Draw below chart.
    \param can Canvas to draw on
    \param loc Chart location to work for */
void
WChartModeList::drawBelow(GfxCanvas& can, GChartLocation& loc)
{
    if (current_mode != 0)
        current_mode->drawBelow(can, loc);
}

/** Draw cursor.
    \param can Canvas to draw on
    \param loc Chart location to work for */
void
WChartModeList::drawCursor(GfxCanvas& can, GChartLocation& loc)
{
    if (current_mode != 0)
        current_mode->drawCursor(can, loc);
}

/** Draw overlays.
    \param can Canvas to draw on
    \param loc Chart location to work for */
void
WChartModeList::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    if (current_mode != 0)
        current_mode->drawOverlays(can, loc);
}

/** Handle event.
    \param event Event to handle
    \param loc Chart location to work for
    \return true if event was handled and main loop need no longer handle it */
bool
WChartModeList::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    return current_mode != 0
        && current_mode->handleEvent(event, loc);
}

/** Mark this mode list changed (will trigger full redraw) */
void
WChartModeList::markChanged()
{
    changed = true;
}

/** Clear change flag. Map will not be (or has just been) redrawn. */
void
WChartModeList::clearChanged()
{
    changed = false;
}

/** Check change flag. */
bool
WChartModeList::isChanged() const
{
    return changed;
}

/** Mark rectangle dirty. At least this rectangle will be redrawn. */
void
WChartModeList::markDirty(GfxRect r)
{
    dirty.include(r);
}

/** Mark clean. Map will not be (or has just been) redrawn. */
void
WChartModeList::markClean()
{
    dirty.w = dirty.h = 0;
}

/** Check for dirty rectangle. */
bool
WChartModeList::isDirty() const
{
    return dirty.exists();
}

/** Get dirty rectangle. */
const GfxRect&
WChartModeList::getDirtyRect() const
{
    return dirty;
}
