/**
  *  \file client/tiles/shipmovement.cc
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/shipmovement.h"
#include "game/mission.h"
#include "game/ship.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "ui/uiglobal.h"
#include "game/formula.h"
#include "game/host.h"
#include "ui/widgets/button.h"
#include "client/tiles/shipmission.h"
#include "game/univ.h"
#include "game/config.h"
#include "game/shippredict.h"

/*************************** WShipMovementTile ***************************/

WShipMovementTile::WShipMovementTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(GfxPoint(315, 7*font_heights[FONT_NORMAL]), selection)
{
    init();
}

void
WShipMovementTile::init()
{
    const int he = 7*font_heights[FONT_NORMAL];
    setLayout(RightJustified, 0, false);
    addControl(GfxRect(17, he - 20, 20, 20), h.add(new UIButton("A", 'a')));
    addControl(GfxRect(17, he - 45, 20, 20), h.add(new UIButton("W", 'w')));

    cf_f10.add(h.add(new UIButton(cm_Ship_FleetMenu, "F10", SDLK_F10)).addCommand());
    addControl(GfxRect( 0, he - 72, 39, 24), cf_f10);
}

void
WShipMovementTile::drawShipMovementTile(UIAbstractTableWriter& tw, bool fixed_only)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Location:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Waypoint:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Distance:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Warp Factor:"));
    tw.writeFixedCell(UIColor::tc_Static, _("E.T.A.:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Fuel Usage:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Engine Load:"));

    if (fixed_only)
        return;                 // save time by not computing movement info

    GShip* ps = dynamic_cast<GShip*>(getCurrentObject());
    GUniverse* univ = getCurrentUniverse();
    if (!ps || ps->getShipKind() != GShip::CurrentShip || !univ)
        return;

    /* compute stuff */
    GShip& s = *ps;
    GShipTurnPredictor crystal_ball(*univ, s.getId());
    crystal_ball.computeMovement();

    const bool is_hyper = s.isHyperdriving();

    tw.newLColumn();

    /* location */
    tw.writeVariableCell(UIColor::tc_Green, univ->getLocationName(s.getPos(), locs_WW | locs_Verbose));

    /* waypoint */
    if (s.getMission() == GMission::msn_Intercept && univ->ty_any_ships.isValidIndex(s.getInterceptId())) {
        tw.writeVariableCell(UIColor::tc_Green, univ->getShip(s.getInterceptId()).getName(GObject::PlainName));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, univ->getLocationName(s.getWaypoint(), locs_WW | locs_Verbose | locs_Ships));
    }

    /* distance */
    double dist = distFromDX(s.getWaypointDX(), s.getWaypointDY());
    tw.writeVariableCell(!is_hyper
                         ? UIColor::tc_Green
                         : dist < 20
                         ? UIColor::tc_Red
                         : !host.isExactHyperjumpDistance2(int32(dist*dist+.1)) /* FIXME: this may lose precision, hence the +.1 */
                         ? UIColor::tc_Yellow
                         : UIColor::tc_Green,
                         format(_("%.2f ly"), dist));

    /* warp */
    if (s.getWarp() == 0 && (s.getWaypointDX() | s.getWaypointDY())) {
        tw.writeVariableCell(UIColor::tc_Red, _("not moving"));
    } else {
        // Borland sucks.
        string_t st;
        if (is_hyper)
            st = _("Hyperdrive");
        else if (s.getWarp() == 0)
            st = _("not moving");
        else
            st = format(_("Warp %d"), s.getWarp());
        tw.writeVariableCell(UIColor::tc_Green, st);
    }

    /* e.t.a. */
    {
        string_t eta;
        UISkin::Color c = UIColor::tc_Green;

        if ((s.getWaypointDX() == 0 && s.getWaypointDY() == 0) && (s.getWarp() > 0 || s.getMission() != GMission::msn_Intercept)) {
            eta = _("at waypoint");
        } else if (is_hyper) {
            eta = _("unknown"); // FIXME?
        } else if (s.getWarp() == 0) {
            c = UIColor::tc_Red;
            eta = _("not moving");
        } else if (crystal_ball.isAtTurnLimit()) {
            eta = _("too long");
        } else {
            eta = format(_("%d turn%!1{s%}"), crystal_ball.getNumTurns());
        }
        tw.writeVariableCell(c, eta);
    }

    /* fuel usage */
    {
        string_t fuel;
        UISkin::Color c = UIColor::tc_Green;

        fuel = format(_("%d kt"), crystal_ball.getMovementFuelUsed());
        if (crystal_ball.isAtTurnLimit()) {
            c = UIColor::tc_Yellow;
        } else if (crystal_ball.getMovementFuelUsed() > s.getCargo(el_Neutronium)
                 || (s.getCargo(el_Neutronium) == 0 && !config.AllowNoFuelMovement())) {
            c = UIColor::tc_Red;
        }

        // FIXME: cloak fuel
        tw.writeVariableCell(c, fuel);
    }

    /* load */
    int towee_mass = 0;
    if (s.getMission() == GMission::msn_Tow && univ->ty_any_ships.isValidIndex(s.getTowId())) {
        mp16_t m = univ->getShip(s.getTowId()).getMass();
        if (m.isKnown()) {
            towee_mass = m;
        }
    }
    tw.writeVariableCell(UIColor::tc_Green, format(_("%d kt"), s.getMass() + towee_mass));
    tw.finish();
}

void
WShipMovementTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(5, 20), FONT_NORMAL);
    drawShipMovementTile(tw, false);

    // F10 button
    TFrameColor color;
    GShip* ps = dynamic_cast<GShip*>(getCurrentObject());
    if (!ps || ps->getFleetNumber() == 0) {
        color = cf_None;
    } else if (ps->isFleetLeader()) {
        color = cf_Green;
    } else {
        color = cf_Red;
    }
    cf_f10.setColor(color);
}
