/**
  *  \file client/winfo.cc
  *  \brief Ship Component Info Windows
  *
  *  This file defines a few widgets that display information about a
  *  starship component.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <vector>
#include "client/winfo.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/spec.h"
#include "gfx/font.h"
#include "ui/skin.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "util/nls.h"
#include "game/host.h"
#include "game/hullfunc.h"
#include "game/exp.h"
#include "game/team.h"
#include "game/playerset.h"

/// Number of lines for hull functions.
enum { HULLINFO_LINES = 10 };

/******************************* WHullInfo *******************************/

/** \class WHullInfo
    \brief Information about a Hull */

/** Constructor. */
WHullInfo::WHullInfo()
    : UIWidget(), hull_id(0), button_s("S", 's')
{
    addChildWidget(button_s, 0);
}

WHullInfo::~WHullInfo()
{ }

/** Set the hull to display.
    \param n Hull number */
void
WHullInfo::setHullId(int n)
{
    if (hull_id != n) {
        hull_id = n;
        drawWidget();
    }
}

void
WHullInfo::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getExtent(), FONT_NORMAL);

    const char* const torp_str    = _("\xE2\x80\xA2 Torpedoes:");
    const char* const fighter_str = _("\xE2\x80\xA2 Fighters:");
    const char* const beam_str    = _("\xE2\x80\xA2 Beams:");

    int min_col_width = std::max(fonts[FONT_NORMAL]->getTextWidth(torp_str),
                                 std::max(fonts[FONT_NORMAL]->getTextWidth(beam_str),
                                          fonts[FONT_NORMAL]->getTextWidth(fighter_str)));

    tw.newLColumn(min_col_width);
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Mass:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Cargo:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Fuel:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Engines:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Crew:"));

    if (hull_id) {
        GHull& h = getHull(hull_id);
        int num_blank = 2; /* engines, crew */

        /* This is a hack - sort of. UITableWriter does not support
           multi-column entries yet. Because it does not expect us to
           write the hull functions as multi-column entries, it will
           not correctly overwrite them and leave dirt around. */
        GfxPoint pt = tw.getCurrentPosition();
        drawBackground(ctx, GfxRect(pt.x, pt.y, extent.x + extent.w - pt.x, extent.y + extent.h - pt.y));

        if (h.getMaxBeams())
            tw.writeVariableCell(UIColor::tc_Static, beam_str), ++num_blank;
        if (h.getMaxLaunchers())
            tw.writeVariableCell(UIColor::tc_Static, torp_str), ++num_blank;
        if (h.getNumBays())
            tw.writeVariableCell(UIColor::tc_Static, fighter_str), ++num_blank;

        tw.newRColumn(4 * font_ems[FONT_NORMAL]);
        tw.writeVariableCell(UIColor::tc_Green, numToString(h.getMass()));
        tw.writeVariableCell(UIColor::tc_Green, numToString(h.getCargo()));
        tw.writeVariableCell(UIColor::tc_Green, numToString(h.getFuel()));
        tw.writeVariableCell(UIColor::tc_Green, numToString(h.getNumEngines()));
        tw.writeVariableCell(UIColor::tc_Green, numToString(h.getCrew()));
        if (h.getMaxBeams())
            tw.writeVariableCell(UIColor::tc_Green, numToString(h.getMaxBeams()));
        if (h.getMaxLaunchers())
            tw.writeVariableCell(UIColor::tc_Green, numToString(h.getMaxLaunchers()));
        if (h.getNumBays())
            tw.writeVariableCell(UIColor::tc_Green, numToString(h.getNumBays()));

        tw.newLColumn();
        tw.writeFixedCell(UIColor::tc_Green, "kt");
        tw.writeFixedCell(UIColor::tc_Green, "kt");
        tw.writeFixedCell(UIColor::tc_Green, "kt");
        do
            tw.writeFixedCell(UIColor::tc_Green, string_t());
        while (--num_blank);
        tw.finish();

        /* build up list of special devices */
        GHullFunctionList funcs;
        h.enumerateHullFunctions(funcs, GPlayerSet(getPlayerId()), GExpLevelSet::allUpTo(NUM_EXPERIENCE_LEVELS), true, false);
        funcs.simplify();
        funcs.sortForNewShip(GPlayerSet(getPlayerId()));

        int line = tw.getCurrentLine();
        ctx.setTextAlign(0, 0);
        ctx.setTransparentBackground();
        ctx.setColor(UIColor::tc_Static);
        ctx.useFont(*fonts[FONT_NORMAL]);

        for (std::size_t i = 0; i < funcs.size() && line < HULLINFO_LINES; ++i) {
            const GBasicHullFunction* ptr = GHullFunctionData::instance.getBasicFunctionById(funcs[i].getBasicFunctionId());
            if (ptr) {
                outTextF(ctx, getExtent().x,
                         getExtent().y + line * font_heights[FONT_NORMAL],
                         getExtent().w,
                         "\xE2\x80\xA2 " + ptr->getDescription());
                ++line;
            }
        }
    }

    can.update(getExtent());
}

void
WHullInfo::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.min_size = li.pref_size = font_sizes[FONT_NORMAL].scaledBy(20, HULLINFO_LINES);
}

void
WHullInfo::onResize(GfxRect /*oldsize*/, bool /*by_layout*/)
{
    button_s.setExtent(GfxRect(getExtent().getRightX() - 20,
                               getExtent().y,
                               20,
                               20),
                       true);
}

/******************************* WBeamInfo *******************************/

/*! \class WBeamInfo
    \brief Information about a beam */

/** Constructor. */
WBeamInfo::WBeamInfo()
    : UIWidget(), beam_id(1)
{ }

WBeamInfo::~WBeamInfo()
{ }

/** Get beam classification. */
static const char*
getBeamType(const GBeam& b)
{
    /* translators: PHost 4 distinguishes between "death rays" and
       normal beams. We invent three new kinds for a change. Invent
       new names if you want. */
    if (host.hasDeathRays() && b.getDamage() == 0)
        return _("death ray");
    else if (b.getDamage() >= 2 * b.getKill())
        return _("atomizer beam");
    else if (b.getKill() >= 2 * b.getDamage())
        return _("X-ray");
    else
        return _("energy beam");
}

void
WBeamInfo::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getExtent(), FONT_NORMAL);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Type:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Kill:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Destroy:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Recharge:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Hit:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Sweep:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Beam Mass:"));

    GBeam& b = getBeam(beam_id);

    tw.newRColumn(3 * font_ems[FONT_NORMAL]);
    GfxPoint mark = tw.getCurrentPosition();
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, numToString(b.getKill()));
    tw.writeVariableCell(UIColor::tc_Green, numToString(b.getDamage()));

    if (host.isTHost()) {
        tw.writeVariableCell(UIColor::tc_Green, "100");
        tw.writeVariableCell(UIColor::tc_Green, "100");
    } else {
        /* Recharge time:
             THost: recharges 0.5 per tick, fires at 50  --> 100
             PHost: recharges by [0, (Kill+Damage) * BeamRechargeBonus + BeamRechargeRate)
               per tick, fires at BeamHitShipCharge
                -> 2*BeamHitShipCharge / ((Kill+Damage) * BRB + BRR)  */
        int bang = b.getKill() + b.getDamage();
        int recharge_rate = bang * config.BeamRechargeBonus(::getPlayerId()) / 100 + config.BeamRechargeRate(::getPlayerId());
        if (!recharge_rate)
            tw.writeVariableCell(UIColor::tc_Green, "-");
        else
            tw.writeVariableCell(UIColor::tc_Green, numToString(2 * config.BeamHitShipCharge(::getPlayerId()) / recharge_rate));

        int hit = bang * config.BeamHitBonus(::getPlayerId()) / 100 + config.BeamHitOdds(::getPlayerId());
        if (hit < 0)
            hit = 0;
        if (hit > 100)
            hit = 100;
        tw.writeVariableCell(UIColor::tc_Green, numToString(hit));
    }

    tw.writeVariableCell(UIColor::tc_Green, numToString(beam_id * beam_id * config.MineSweepRate(::getPlayerId())));
    tw.writeVariableCell(UIColor::tc_Green, numToString(b.getMass()));

    tw.newLColumn();
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, "s");
    tw.writeVariableCell(UIColor::tc_Green, "%");
    if (config.WebMineSweepRate(::getPlayerId()) == config.MineSweepRate(::getPlayerId()))
        tw.writeVariableCell(UIColor::tc_Green, _("mines"));
    else
        tw.writeVariableCell(UIColor::tc_Green, format(_("mines, %d webs"),
                                                       beam_id * beam_id * config.WebMineSweepRate(::getPlayerId())));
    tw.writeVariableCell(UIColor::tc_Green, "kt");
    tw.finish();

    ctx.setSolidBackground();
    ctx.setTextAlign(0, 0);
    ctx.setColor(UIColor::tc_Green);
    ctx.useFont(*fonts[FONT_NORMAL]);
    outTextF(ctx, mark.x, mark.y, getExtent().x + getExtent().w - mark.x, getBeamType(b));
    can.update(getExtent());
}

void
WBeamInfo::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.min_size = li.pref_size = font_sizes[FONT_NORMAL].scaledBy(20, 7);
}

void
WBeamInfo::setBeamId(int n)
{
    if (n != beam_id) {
        beam_id = n;
        drawWidget();
    }
}

/******************************* WTorpInfo *******************************/

/*! \class WTorpInfo
    \brief Information about a torpedo */
WTorpInfo::WTorpInfo()
    : UIWidget(), torp_id(1)
{ }

WTorpInfo::~WTorpInfo()
{ }

/** Get torpedo classification. */
static const char*
getTorpType(const GTorpLauncher& b)
{
    /* translators: PHost 4 distinguishes between "death rays" and
       normal torpedoes. We invent three new kinds for a change. 
       Invent new names if you want. */
    if (host.hasDeathRays() && b.getDamage() == 0)
        return _("death ray torp");
    else if (b.getDamage() >= 2 * b.getKill())
        return _("atomizer torp");
    else if (b.getKill() >= 2 * b.getDamage())
        return _("X-ray torp");
    else
        return _("explosive torp");
}

void
WTorpInfo::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getExtent(), FONT_NORMAL);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Type:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Kill:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Destroy:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Recharge:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Hit:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Torp Cost:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 1000 Mines:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Launcher Mass:"));

    GTorpLauncher& l = getTorp(torp_id);
    int scale = (host.isTHost() || !config.AllowAlternativeCombat()) ? 2 : 1;

    tw.newRColumn(3 * font_ems[FONT_NORMAL]);
    GfxPoint mark = tw.getCurrentPosition();
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, numToString(scale * l.getKill()));
    tw.writeVariableCell(UIColor::tc_Green, numToString(scale * l.getDamage()));

    if (host.isTHost()) {
        /* These values directly follow from the Tim-VCR code,
           although they are never explicitly used as such. */
        static const int8 recharge_times[] = {
           41, 41, 36, 34, 33, 33, 32, 32, 32, 31
        };
        tw.writeVariableCell(UIColor::tc_Green, numToString(recharge_times[torp_id-1]));
        tw.writeVariableCell(UIColor::tc_Green, "66");
    } else {
        int bang = l.getKill() + l.getDamage();
        int rate = bang * config.TubeRechargeBonus(::getPlayerId()) / 100 + config.TubeRechargeRate(::getPlayerId());
        if (rate)
            tw.writeVariableCell(UIColor::tc_Green, numToString(2000 / rate));
        else
            tw.writeVariableCell(UIColor::tc_Green, "-");
        rate = bang * config.TorpHitBonus(::getPlayerId()) / 100 + config.TorpHitOdds(::getPlayerId());
        if (rate < 0)
            rate = 0;
        if (rate > 100)
            rate = 100;
        tw.writeVariableCell(UIColor::tc_Green, numToString(rate));
    }

    tw.writeVariableCell(UIColor::tc_Green, numToString(l.getTorpCost()));

    /* how many mines do we get from 100 torps? */
    int mine_lay_divi = (host.isPHost()
                         ? config.UnitsPerTorpRate(::getPlayerId())
                         : raceId(::getPlayerId()) == 9
                         ? 400 : 100)
        * torp_id * torp_id;

    /* simple equation: if 100 torps give mine_lay_divi mines, how
       many torps do we need for 1000 mines? */
    if (mine_lay_divi <= 0)
        tw.writeVariableCell(UIColor::tc_Green, "-");
    else
        tw.writeVariableCell(UIColor::tc_Green, numToString(100000L * l.getTorpCost() / mine_lay_divi));
    tw.writeVariableCell(UIColor::tc_Green, numToString(l.getMass()));

    // FIXME: this assumes 1 TDM cost for torpedoes. Our interface would allow different
    // costs, we should probably be able to handle that.
    tw.newLColumn();
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, "s");
    tw.writeVariableCell(UIColor::tc_Green, "%");
    tw.writeVariableCell(UIColor::tc_Green, _("mc, 1 T/D/M"));
    tw.writeVariableCell(UIColor::tc_Green, format("mc, %d T/D/M", 100000 / mine_lay_divi));
    tw.writeVariableCell(UIColor::tc_Green, "kt");
    tw.finish();

    ctx.setSolidBackground();
    ctx.setTextAlign(0, 0);
    ctx.setColor(UIColor::tc_Green);
    ctx.useFont(*fonts[FONT_NORMAL]);
    outTextF(ctx, mark.x, mark.y, getExtent().x + getExtent().w - mark.x, getTorpType(l));
    can.update(getExtent());
}

void
WTorpInfo::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.min_size = li.pref_size = font_sizes[FONT_NORMAL].scaledBy(20, 8);
}

void
WTorpInfo::setTorpId(int n)
{
    if (n != torp_id) {
        torp_id = n;
        drawWidget();
    }
}

/****************************** WEngineInfo ******************************/

WEngineInfo::WEngineInfo()
    : UIWidget(), engine_id(1)
{ }

WEngineInfo::~WEngineInfo()
{ }

void
WEngineInfo::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getExtent(), FONT_NORMAL);
    GEngine& e = getEngine(engine_id);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Max efficient Warp:"));
    if (config.AllowEngineShieldBonus())
        tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Shield Bonus:"));

    tw.newRColumn(3 * font_ems[FONT_NORMAL]);
    tw.writeVariableCell(UIColor::tc_Green, numToString(e.getMaxEfficientWarp()));
    if (config.AllowEngineShieldBonus())
        tw.writeVariableCell(UIColor::tc_Green,
                             numToString(e.getCost().get(el_Money) * config.EngineShieldBonusRate(::getPlayerId()) / 100));

    tw.newLColumn();
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    if (config.AllowEngineShieldBonus())
        tw.writeFixedCell(UIColor::tc_Green, "kt");

    tw.finish();
    can.update(getExtent());
}

void
WEngineInfo::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.min_size = li.pref_size = font_sizes[FONT_NORMAL].scaledBy(20, 2);
}

void
WEngineInfo::setEngineId(int n)
{
    if (n != engine_id) {
        engine_id = n;
        drawWidget();
    }
}
