/**
  *  \file game/sim-result.h
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_SIM_RESULT_H
#define PCC_V2_GAME_SIM_RESULT_H

#include "game/sim-run.h"
#include "game/playerset.h"
#include "cpluslib/ptrvector.h"

/*! \addtogroup Simulator */
/*@{*/

/** Statistics counter. Counts minimum, maximum and total (for average
    computation). */
class GSimStatItem {
    int32  min;
    int32  max;
    int32  total_scaled;
 public:
    /** Make blank result. */
    GSimStatItem()
        : min(0), max(0), total_scaled(0)
        { }

    /** Make inverted result. Make result that would have been
        obtained if each call to add(x,w) would have been replaced by
        add(subtract_from-x,w).
        \param orig          [in] original object
        \param subtract_from [in] value to subtract this from
        \param scale         [in] total scale (sum of all w.this_battle_weight used in making this result) */
    GSimStatItem(const GSimStatItem& orig, int32 subtract_from, int32 scale)
        : min(subtract_from - orig.max),
          max(subtract_from - orig.min),
          total_scaled(subtract_from * scale - orig.total_scaled)
        { }
    
    void add(int32 value, const GSimBattleResult& w);

    /** Get minimum. */
    int32 getMin() const
        { return min; }
    /** Get maximum. */
    int32 getMax() const
        { return max; }
    /** Get scaled total. Must be divided by sum of all
        w.this_battle_weight used in making this result to obtain
        average. */
    int32 getTotalScaled() const
        { return total_scaled; }
    /** Change weight proportionally. Assuming this result was
        obtained using weight \c old_weight, modify it so that it is
        appropriate for \c new_weight. */
    void changeWeight(int32 old_weight, int32 new_weight)
        { total_scaled = total_scaled * new_weight / old_weight; }
};

/** Summary of a sequence of simulations. Contains statistics for each
    unit (UnitResult) as well as different result classes including
    example battles (ClassResult).

    Every battle can have a weight. For example, with Host's
    left/right balancing, one particular setup with one particular
    seed has a 59-vs-41% weight of occuring. We do not simulate these
    59+41 battles, but instead only simulate two and adjust their
    weights accordingly. For efficiency, only the GSimResultSummary
    stores the effective total weight, the GSimStatItem's do not know
    the value internally. */
class GSimResultSummary {
 public:
    /** Overall simulation result for a single unit. Statistics counters. */
    struct UnitResult {
        /** Number of times this ship survived. */
        int won;
        /** Number of times this ship fought. */
        int fought;
        /** Number of times this ship got captured (survived with
            different owner). */
        int captured;
        /** Number of torpedoes fired (ships and planets). */
        GSimStatItem torps_fired;
        /** Number of fighters lost (ships and planets). */
        GSimStatItem fighters_lost;
        /** Damage taken (ships and planets). */
        GSimStatItem damage_taken;
        /** Shields left (ships and planets). */
        GSimStatItem shields_left;
        /** Crew left (ships) or defense lost (planets). */
        GSimStatItem crew_left_or_defense_lost;
        /** Torps hit (ships and planets). */
        GSimStatItem torps_hit;
        /** Minimum fighters on unit at any one time (ships and
            planets). */
        GSimStatItem min_fighters_aboard;
        
        UnitResult()
            : won(0), fought(0), captured(0)
            { }
        void changeWeight(int32 old_weight, int32 new_weight);
        void addResult(const GSimShip& old_ship, const GSimShip& new_ship, const GSimBattleResult& res);
        void addResult(const GSimPlanet& old_planet, const GSimPlanet& new_planet, const GSimBattleResult& res);
    };

    /** Result summary grouped by result. Classes are defined by
        having similar result sets; this object contains information
        about one such class. Similarity is so far defined by
        identical numbers of surviving ships. */
    struct ClassResult {
        /** Number of units owned by this player. */
        GPlayerArray<int> owned_units;
        /** Cumulative weight of this result. */
        int32 cumulative_weight;
        /** An example battle. */
        Ptr<GVcrDatabase> sample_battle;

        ClassResult(const GSimState& new_state, const GSimBattleResult& result);
        ~ClassResult();

        void changeWeight(int32 old_weight, int32 new_weight)
            { cumulative_weight = cumulative_weight * new_weight / old_weight; }

        bool isSameClass(const ClassResult& other) const;
        void addNewSameClassResult(const ClassResult& other);
    };

    GSimResultSummary();
    ~GSimResultSummary();
    void addResult(const GSimState& old_state, const GSimState& new_state, GSimBattleResult result);

    /** Get cumulative weight. */
    int32 getCumulativeWeight() const
        { return cumulative_weight; }
    /** Get total weight to which weights are normalized. */
    int32 getTotalWeight() const
        { return total_weight; }
    /** Get number of result classes. */
    int getNumClassResults() const
        { return class_results.size(); }
    /** Get number of unit results. */
    int getNumUnitResults() const
        { return unit_results.size(); }
    /** Get one class result. */
    const ClassResult& getClassResult(int index) const
        { return *class_results[index]; }
    /** Get one unit result. The index is in range [0,N) for a N-ship
        battle, N is used for the planet if applicable. This makes
        indexes compatible with WSimList. */
    const UnitResult& getUnitResult(int index) const
        { return *unit_results[index]; }
    /** Return number of battles already fought. This is the number of
        addResult() invocations. */
    int getNumBattles() const
        { return battle_count; }

    bool isPlanet(int index) const
        { return index >= 0 && index == planet_index; }

 private:
    /** Total weight. This is the value to which the battles are
        "normalized". */
    int32 total_weight;
    /** Cumulative weight of all fights. */
    int32 cumulative_weight;
    /** Total number of battles so far. */
    int battle_count;
    /** Index of planet in unit_results. -1 if none. */
    int planet_index;

    typedef ptr_vector<UnitResult> unitvec_t;
    typedef ptr_vector<ClassResult> classvec_t;
    
    unitvec_t  unit_results;
    classvec_t class_results;

    void updateClassResultSortOrder(classvec_t::size_type change_index);
};
/*@}*/

#endif
