/**
  *  \file vcr.cc
  *  \brief VCR Player Interface
  *
  *  The abstract interface for playing battles. This defines the
  *  interface used by players and visualizers.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/vcr.h"
#include "game/config.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/team.h"
#include "util/recode.h"
#include "io/io.h"

/****************************** GClassicVcr ******************************/

GClassicVcr::GClassicVcr()
    : seed(0),
      signature(0),
      planet_temp_code(0)
{ }

GClassicVcr::GClassicVcr(const TRawVcr& raw)
    : seed(raw.random_seed),
      signature(raw.signature),
      planet_temp_code(raw._planet_temp_code)
{
    sides[0].unpack(raw, 0);
    sides[1].unpack(raw, 1);
}

GClassicVcr::~GClassicVcr()
{ }

/******************************* VcrPlayer *******************************/

/*! \class VcrPlayer

    This class is the base for all VCR player algorithms. Its
    interface includes saving and restoring state, and, of course,
    playing the VCR. In addition, it has an interface to gather some
    statistical information for simulation, which cannot easily
    gathered from the outside.

    Playing interface:
    - initVcr(vcr) initializes the player for the given VCR.
      This includes a call to the visualizer's init();
    - optionally call playFastForward();
    - call playCycle() repeatedly. If playCycle() returns true,
      the battle continues. If playCycle() returns false,
      the battle has ended, and playCycle() did nothing.
      The first thing playCycle() does is incrementing the
      timer.
    - doneVcr(res) finishes playing and determines the result.
      This does final exploding and such.

    playVcr(vcr) does all but the last step at once.

    Before calling the playing interface, you must associate
    a visualizer with the VcrPlayer. */

/** Constructor. Creates an empty VCR player not associated with
    a viewer. */
VcrPlayer::VcrPlayer()
{
    vis = 0;
}

VcrPlayer::~VcrPlayer()
{
    if (vis.ptr())
        vis->setPeer(0);
    vis = 0;
}

/** Play VCR. Shortcut for playing a whole VCR at once.
    Does not necessarily keep display up-to-date. */
void
VcrPlayer::playVcr(const GClassicVcr& vcr, ProgressMonitor& pm)
{
    initVcr(vcr);
    playFastForward();
    int tick = 0;
    while (playCycle()) {
        if (++tick > 100) {
            pm.tick();
            tick = 0;
        }
    }
}

/** Set Visualizer. Before anything goes here, there must be
    a visualizer connected to the player.
    \param avis new visualizer. Can be null to remove the
    current association. */
void
VcrPlayer::setVisualizer(Ptr<VcrVisualizer> avis)
{
    if (vis.ptr())
        vis->setPeer(0);
    vis = avis;
    if (vis.ptr())
        vis->setPeer(this);
}

/** Reset statistics. Must be called by initVcr(), before the first
    playCycle() / playFastForward(). */
void
VcrPlayer::initStat()
{
    ASSERT(getTime() == 0);
    for (int side = 0; side < 2; ++side) {
        stat_info[side].torps_hit = 0;
        stat_info[side].min_fighters_aboard = getFighterCount(side);
        stat_info[side].num_fights = 1;
    }
}

/** Fast forward. If possible, jump forward in time.
    Does not need to keep the display up-to-date.
    If the VCR algorithm supports that, override this
    method. \default does nothing. */
void
VcrPlayer::playFastForward()
{ }

/** Format as string. \returns human-readable description of result status
    for battle vcr. */
string_t
VcrPlayer::getResultString(int status, const GClassicVcr& vcr)
{
    int me;
    if (vcr(0).getOwner() == getPlayerId())
        me = 0;
    else if (vcr(1).getOwner() == getPlayerId())
        me = 1;
    else
        me = 2;

    string_t s;
    switch (status) {
     case 0:
        s = _("unknown. Wait while computing...");
        break;
     case VcrPlayer::VCRS_INVALID:
        s = _("Battle cannot be played!");
        break;
     case VcrPlayer::VCRS_TIMEOUT:
        s = _("Battle timed out (too long).");
        break;
     case VcrPlayer::VCRS_STALEMATE:
        s = _("Stalemate.");
        break;
     case VcrPlayer::VCRS_LEFT_DESTROYED:
        if (me == 0)
            s = _("We were destroyed.");
        else if (me == 1)
            s = _("We won.");
        else
            s = format(_("%s won."), vcr(1).getName());
        break;
     case VcrPlayer::VCRS_RIGHT_DESTROYED:
        if (me == 1)
            s = _("We were destroyed.");
        else if (me == 0)
            s = _("We won.");
        else
            s = format(_("%s won."), vcr(0).getName());
        break;
     case VcrPlayer::VCRS_LEFT_CAPTURED:
        if (me == 0)
            s = _("They have captured our ship.");
        else if (me == 1)
            s = _("We captured their ship.");
        else
            s = format(_("%s was captured."), vcr(0).getName());
        break;
     case VcrPlayer::VCRS_RIGHT_CAPTURED:
        if (me == 1)
            s = _("They have captured our ship.");
        else if (me == 0)
            s = _("We captured their ship.");
        else
            s = format(_("%s was captured."), vcr(1).getName());
        break;
     case VcrPlayer::VCRS_RIGHT_DESTROYED + VcrPlayer::VCRS_LEFT_DESTROYED:
        s = _("Both were destroyed.");
        break;
     default:
        s = _("Both are disabled.");
    }
    return s;
}

/*! \class VcrStatus
    VCR Status token. VcrPlayer descendants derive a class from this
    and store information they need in here. */
VcrStatus::VcrStatus(VcrPlayer::time_type time)
    : time (time)
{ }

VcrStatus::~VcrStatus()
{ }

/*! \class VcrVisualizer

    The visualizer's methods are called by a VCR player to do actual
    screen display. Not all events generate visualizer callbacks, hence
    the visualisation code must also poll for changes (fighter positions,
    damage/kill values, ...). It can use the accessor interface of
    VcrPlayer for that. */
VcrVisualizer::VcrVisualizer()
    : ply(0)
{ }

VcrVisualizer::~VcrVisualizer()
{ }

/** Set Player. This method is called by the VcrPlayer to inform
    this visualizer whom it works for. */
void
VcrVisualizer::setPeer(VcrPlayer* aply)
{
    ply = aply;
}

/** \class VcrNullVisualizer
    \brief Null Visualizer

    This VCR visualizer does nothing. It is to be used for
    simulations, etc. */
void
VcrNullVisualizer::startFighter(int /*side*/, int /*track*/)
{ }

void
VcrNullVisualizer::landFighter(int /*side*/, int /*track*/)
{ }

void
VcrNullVisualizer::killFighter(int /*side*/, int /*track*/)
{ }

void
VcrNullVisualizer::fireBeam(int /*side*/, int /*track*/, int /*target*/, int /*hit*/, int, int)
{ }

void
VcrNullVisualizer::fireTorp(int /*side*/, int /*hit*/, int)
{ }

void
VcrNullVisualizer::updateBeam(int /*side*/, int /*id*/)
{ }

void
VcrNullVisualizer::updateTorp(int /*side*/, int /*id*/)
{ }

void
VcrNullVisualizer::killObject(int /*side*/)
{ }

void
VcrNullVisualizer::redraw()
{ }

void
VcrNullVisualizer::init()
{ }
