/**
  *  \file io/zipstream.cc
  *  \brief Zip (deflate) stream
  *
  *  Format of gzip stream:
  *    BYTE     magic 1, 0x1F
  *    BYTE     magic 2, 0x8B
  *    BYTE     compression method, 8
  *    BYTE     flag
  *              bit 0    ASCII (not binary)
  *              bit 1    header CRC present
  *              bit 2    extra data present
  *              bit 3    file name present
  *              bit 4    comment present
  *    DWORD    mtime
  *    BYTE     flag
  *    BYTE     OS
  *    --- if flag bit 2 set ---
  *    WORD     length
  *    BYTE[]   data
  *    --- if flag bit 3 set ---
  *    BYTE[]   filename, 0-terminated
  *    --- if flag bit 4 set ---
  *    BYTE[]   comment, 0-terminated
  *    --- if flag bit 1 set ---
  *    WORD     CRC
  *    --- always ---
  *    BYTE[]   compressed stream
  *    DWORD    CRC32
  *    DWORD    size
  *  CRC and size are NOT verified.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "io/zipstream.h"
#include "util/except.h"
#include "util/nls.h"

namespace {
    enum {
        flASCII     = 1,
        flHeaderCRC = 2,
        flExtraData = 4,
        flFileName  = 8,
        flComment   = 16,
        flAll       = flASCII + flHeaderCRC + flExtraData + flFileName + flComment
    };
};

/** Initialize.
    \param worker Underlying stream
    \param gzip true to read gzip format, false to read raw deflate format */
ZipStream::ZipStream(Stream& worker, bool gzip)
    : worker(worker),
      outPtr(0),
      bytesRead(0),
      state(gzip ? CheckHeader : FinishedHeader),
      endReached(false)
{
    zs.next_in   = 0;
    zs.avail_in  = 0;
    zs.total_in  = 0;
    zs.next_out  = 0;
    zs.avail_out = 0;
    zs.total_out = 0;
    zs.zalloc    = 0; /* use default */
    zs.zfree     = 0; /* use default */
    zs.opaque    = this;
}

ZipStream::~ZipStream()
{
    if (state == Decompressing) {
        inflateEnd(&zs);
    }
}

/** Throw exception corresponding to a zlib error code. */
void
ZipStream::zlibError(int state)
{
    /* zError and .msg return constant ASCII strings, so no codepage conversion needed here. */
    string_t message = _("Decompression failed: ");
    message += zError(state);
    if (zs.msg) {
        message += ", ";
        message += zs.msg;
    }
    throw FileIOException(*this, message);
}

/** Refill input buffer. A partial leftover will be kept. */
void
ZipStream::refillInput()
{
    /* Only refill the buffer if it has room and we haven't yet reached the end */
    if (zs.avail_in < sizeof(bufIn) && !endReached) {
        /* Keep unread input, if any */
        std::size_t sizeToRead = sizeof(bufIn);
        if (zs.next_in != 0) {
            memmove(bufIn, zs.next_in, zs.avail_in);
            sizeToRead -= zs.avail_in;
        }

        /* Read new input */
        zs.next_in = reinterpret_cast<Bytef*>(bufIn);
        zs.avail_in += worker.read(reinterpret_cast<char*>(zs.next_in) + zs.avail_in, sizeToRead);
    }
}

/** Consume some input bytes. */
void
ZipStream::consumeInput(std::size_t n)
{
    zs.avail_in -= n;
    zs.next_in += n;
}

/** Consume a string.
    \retval true String finished
    \retval false String not finished */
bool
ZipStream::consumeString()
{
    /* Read until end of string */
    while (zs.avail_in > 0 && zs.next_in[0] != 0) {
        consumeInput(1);
    }
    /* End reached? */
    if (zs.avail_in == 0) {
        return false;
    }
    consumeInput(1);
    return true;
}

/** Process header. Simple state machine.
    \retval true Some data successfully processed
    \retval false No data processed, must read some more data */
bool
ZipStream::processHeader()
{
    switch (state) {
     case CheckHeader:
        /* Must have at least 10 bytes */
        if (zs.avail_in < 10) {
            return false;
        }

        /* Check magic, compression method, and flags */
        if (uint8_t(zs.next_in[0]) != 0x1F
            || uint8_t(zs.next_in[1]) != 0x8B
            || uint8_t(zs.next_in[2]) != 8
            || uint8_t(zs.next_in[3] & ~flAll) != 0)
        {
            throw FileFormatException(*this, _("Invalid file header"));
        }

        /* Remember data */
        gzHeaderFlags = zs.next_in[3];

        /* Advance state */
        consumeInput(10);
        state = CheckExtraDataLength;
        return true;

     case CheckExtraDataLength:
        /* Read length if requested */
        if (gzHeaderFlags & flExtraData) {
            /* Must have two bytes */
            if (zs.avail_in < 2) {
                return false;
            }
            gzExtraLength = uint8_t(zs.next_in[0]) + 256*uint8_t(zs.next_in[1]);
            consumeInput(2);
            state = CheckExtraData;
        } else {
            /* No header */
            state = CheckFileName;
        }
        return true;

     case CheckExtraData:
        /* Read data */
        if (zs.avail_in == 0) {
            return false;
        }
        if (zs.avail_in >= gzExtraLength) {
            consumeInput(gzExtraLength);
            state = CheckFileName;
        } else {
            gzExtraLength -= zs.avail_in;
            consumeInput(zs.avail_in);
        }
        return true;

     case CheckFileName:
        /* Read null-terminated file name */
        if (gzHeaderFlags & flFileName) {
            if (zs.avail_in == 0) {
                return false;
            }
            if (!consumeString()) {
                return true;
            }
        }
        state = CheckComment;
        return true;

     case CheckComment:
        /* Read null-terminated comment */
        if (gzHeaderFlags & flComment) {
            if (zs.avail_in == 0) {
                return false;
            }
            if (!consumeString()) {
                return true;
            }
        }
        state = CheckHeaderCRC;
        return true;

     case CheckHeaderCRC:
        /* Read header CRC */
        if (gzHeaderFlags & flHeaderCRC) {
            if (zs.avail_in < 2) {
                return false;
            }
            consumeInput(2);
        }
        state = FinishedHeader;
        return true;

     case FinishedHeader:
        /* Final state: initialize zlib without consuming more data */
        int result;
        result = inflateInit2(&zs, -15);
        if (result != Z_OK) {
            zlibError(result);
        }
        state = Decompressing;
        return true;

     case Decompressing:
        return true;
    }

    return false;
}

std::size_t
ZipStream::read(char* buffer, std::size_t max)
{
    /* Process headers */
    while (state != Decompressing) {
        if (!processHeader()) {
            refillInput();
            if (!processHeader()) {
                return 0;
            }
        }
    }

    /* Read data */
    std::size_t did = 0;
    while (did < max) {
        /* Try to consume pending output data */
        std::size_t remain = max - did;
        if (outPtr != 0 && zs.next_out != outPtr) {
            /* There is some data from previous decompression call */
            std::size_t avail = zs.next_out - outPtr;
            std::size_t now = avail > remain ? remain : avail;
            std::memcpy(&buffer[did], outPtr, now);
            did += now;
            outPtr += now;
        } else if (endReached) {
            /* End reached */
            break;
        } else {
            /* No data remains in output buffer. Refill input and inflate it */
            refillInput();
            bool anyInput = (zs.avail_in != 0);
            zs.avail_out = sizeof(bufOut);
            zs.next_out = outPtr = reinterpret_cast<Bytef*>(bufOut);
            int result = inflate(&zs, 0);
            if (result == Z_STREAM_END || (result == Z_BUF_ERROR && !anyInput)) {
                /* Stream end marker found, or stream ends and inflate() did
                   not produce any buffered output */
                endReached = true;
            } else {
                /* Success, or a real error */
                if (result != Z_OK) {
                    zlibError(result);
                }
            }
        }
    }

    bytesRead += did;
    return did;
}

std::size_t
ZipStream::write(const char* /*buffer*/, std::size_t /*max*/)
{
    return notWritable();
}

int
ZipStream::getCapabilities() const
{
    return C_READ;
}

void
ZipStream::seek(fpos_t pos)
{
    if (pos != bytesRead) {
        notSeekable();
    }
}

ZipStream::fpos_t
ZipStream::getPos()
{
    return bytesRead;
}

ZipStream::fpos_t
ZipStream::getSize()
{
    return -1;
}

string_t
ZipStream::getName()
{
    return worker.getName();
}
